import os, datetime, subprocess, logging
from scipy.io import netcdf
import numpy as np
import identifiers.poloidal_plane_coordinates_identifier
import imas
import aug_sfutils as sf
import trview
from trview import rhop2rhot

logger = logging.getLogger('trview.trv2imas')


spec_lbl = {(1, 1): 'H', (1, 2): 'D', (2, 4): 'He', (6, 12): 'C'}


def TRV2IMAS(gpar, ion_d, equ, profiles, sig, nbi, ech, icr, write_ids=False):

    logger.info('Creating IDS structure')
#        self.obj = imas.ids(gpar.shot, gpar.IDSrun, 0, 0)
#        self.obj.create_env(os.getenv('USER'), 'aug', '3')
#        print(imas.imasdef.__dict__.keys())
    db = imas.DBEntry(imas.imasdef.ASCII_BACKEND,'aug', gpar.shot, gpar.IDSrun, os.getenv('USER'), '3')
    status, _ = db.create()

    logger.info('Filling IDS structure')
    ds = fill_ds(gpar.shot)
    wall = fill_wall(gpar.shot)
    tf = fill_tf(equ.time, equ.R0, equ.B0)
    if nbi.status and np.max(nbi.pnbi_red) > 0:
        nb = fill_nbi(nbi)
    if hasattr(ech, 't_ds'):
        ec = fill_ech(ech)
    if hasattr(icr, 't_icp'):
        ic = fill_icrf(icr)
    eq = fill_equ(equ)
    ps = fill_ps(nbi, ech, icr)
# Core profiles
    cp = fill_coreprof(equ, profiles, sig, ion_d)

    if write_ids:
        logger.info('Dumping IDS file')
        db.put(ds)
        logger.info('Putting tf')
        db.put(tf)
        logger.info('Putting wall')
        db.put(wall)
        if nb.ids_properties.homogeneous_time >= 0:
            logger.info('Putting NBI object')
            db.put(nb)
        if ec.ids_properties.homogeneous_time >= 0:
            logger.info('Putting ECH object')
            db.put(ec)
        if ic.ids_properties.homogeneous_time >= 0:
            logger.info('Putting ICRF object')
            db.put(ic)
        logger.info('Putting equilibrium')
        db.put(eq)
        logger.info('Putting core_profiles')
        db.put(cp)
        logger.info('Putting pulse_schedule')
        db.put(ps)

        logger.info('Closed IMAS file')


def fill_ds(shot):

    ds = imas.dataset_description()
    ds.ids_properties.homogeneous_time = 2
    ds.ids_properties.comment = "trview (https://www.aug.ipp.mpg.de/aug/manuals/transp/trview, version %s)" %trview.__version__
    ds.ids_properties.source = os.path.abspath(__file__)
    ds.ids_properties.provider = "Giovanni Tardini (giovanni.tardini@ipp.mpg.de)"
    ds.ids_properties.creation_date = datetime.datetime.today().strftime("%Y/%m/%d")
    ds.data_entry.user = os.getenv('USER')
    ds.data_entry.machine = "ASDEX Upgrade"
    ds.data_entry.pulse_type = "pulse"
    ds.data_entry.pulse = shot

    return ds


def fill_wall(shot):

    wall = imas.wall()
    gc_d = sf.getgc(shot)

    wall.ids_properties.comment = 'Based on YGC'
    wall.ids_properties.homogeneous_time = 1
    wall.ids_properties.source = 'AUG Shotfile'
    wall.ids_properties.provider = os.getenv("USER")
    wall.ids_properties.creation_date = datetime.datetime.today().strftime("%Y-%m-%d")
    wall.code.name= "getgc"
    wall.output_flag = 0
    wall.time = np.array([0.0])
    wall.description_2d.resize(2)

    wall.description_2d[0].type.name = 'wall'
    wall.description_2d[0].type.index = 1    # for gas-tight walls (disjoint PFCs with inner vessel as last limiter_unit; no vessel structure);
    wall.description_2d[0].type.description = 'AUG wall (gast-tight)'
    wall.description_2d[0].limiter.unit.resize(1)
    for index, key in enumerate(gc_d.keys()):
        gcr = np.array(gc_d[key].r)
        gcz = np.array(gc_d[key].z)

        if (gcr.min() < 1.65) & ( gcr.max() > 1.65) & (gcz.min() < 0.00) & ( gcz.max() > 0.00):

            wall.description_2d[0].limiter.unit[0].name = key
            if (gcr[0] == gcr[-1]) & (gcz[0] == gcz[-1]):
                wall.description_2d[0].limiter.unit[0].closed = 1
                wall.description_2d[0].limiter.unit[0].outline.r = gcr[:-1]
                wall.description_2d[0].limiter.unit[0].outline.z = gcz[:-1]
            else:
                wall.description_2d[0].limiter.unit[0].closed = 0
                wall.description_2d[0].limiter.unit[0].outline.r = gcr
                wall.description_2d[0].limiter.unit[0].outline.z = gcz

    wall.description_2d[1].type.name = 'wall'
    wall.description_2d[1].type.index = 2    # for free boundary codes (disjoint PFCs and vessel) 
    wall.description_2d[1].type.description = 'AUG wall (complete)'
    wall.description_2d[1].limiter.unit.resize(len(gc_d.keys()))
    for index, key in enumerate(gc_d.keys()):
        gcr = np.array(gc_d[key].r)
        gcz = np.array(gc_d[key].z)
        wall.description_2d[1].limiter.unit[index].name = key
        if (gcr[0] == gcr[-1]) & (gcz[0] == gcz[-1]):
            wall.description_2d[1].limiter.unit[index].closed = 1
            wall.description_2d[1].limiter.unit[index].outline.r = gcr[:-1]
            wall.description_2d[1].limiter.unit[index].outline.z = gcz[:-1]
        else:
            wall.description_2d[1].limiter.unit[index].closed = 0
            wall.description_2d[1].limiter.unit[index].outline.r = gcr
            wall.description_2d[1].limiter.unit[index].outline.z = gcz

    return wall


def fill_tf(tim, R0, B0):

    tf = imas.tf()
    tf.ids_properties.comment = 'Based on jpol from EQI'
    tf.ids_properties.homogeneous_time = 1
    tf.ids_properties.source = 'AUG Shotfile'
    tf.ids_properties.provider = os.getenv("USER")
    tf.ids_properties.creation_date = datetime.datetime.today().strftime("%Y-%m-%d")

    tf.r0 = R0
    B0 = np.array(B0)

    tf.b_field_tor_vacuum_r.data = B0 * R0
    tf.time = np.array(tim)

    return tf


def fill_nbi(nbi):

    nb = imas.nbi()
    nb.ids_properties.creation_date = datetime.datetime.today().strftime("%d/%m/%y")
    nb.ids_properties.homogeneous_time = 0
    nb.ids_properties.source = os.path.abspath(__file__)

    n_nbi = np.sum(nbi.n_box)
    nb.unit.resize(n_nbi)
    tim = np.atleast_1d(np.average(nbi.t_ds))
    Rsrc = np.hypot(nbi.xsrc, nbi.ysrc)
    zsrc = nbi.xybsca
    phi_src = np.arctan2(nbi.ysrc, nbi.xsrc)
    logger.info('Rsrc1 = %8.4f m, zsrc1 = %8.4f m, phisrc1 = %8.4f rad', Rsrc[0], zsrc[0], phi_src[0])
    for jnb in range(n_nbi):
        pow_frac = nbi.power_mix[jnb, :].reshape((3, 1)) # energy_mix, t
        part_frac = nbi.part_mix[jnb, :].reshape((3, 1)) # energy_mix, t
        if jnb < 4:
            nb.unit[jnb].name = 'INJ1-%s' % (jnb+1)
        else:
            nb.unit[jnb].name = 'INJ2-%s' % (jnb-3)
        nb.unit[jnb].species.a   = np.float64(nbi.abeam[jnb])
        nb.unit[jnb].species.z_n = np.float64(nbi.zbeam[jnb])
        nb.unit[jnb].species.label = spec_lbl[(nbi.zbeam[jnb], nbi.abeam[jnb])]
        nb.unit[jnb].beam_current_fraction.time = tim
        nb.unit[jnb].beam_current_fraction.data = part_frac
        nb.unit[jnb].beam_power_fraction.time = tim
        nb.unit[jnb].beam_power_fraction.data = pow_frac
        nb.unit[jnb].energy.time = tim
        nb.unit[jnb].energy.data = np.atleast_1d(1e3*nbi.einj_kev[jnb]) #keV->eV
# Power: time dependent
        nb.unit[jnb].power_launched.time = nbi.t_ds
        nb.unit[jnb].power_launched.data = nbi.pnbi_ds[:, jnb]
        nb.unit[jnb].beamlets_group.resize(1)
        nb.unit[jnb].beamlets_group[0].tangency_radius = 1e-2*nbi.rtcena[jnb]
        nb.unit[jnb].beamlets_group[0].position.r   = 1e-2*Rsrc[jnb]
        nb.unit[jnb].beamlets_group[0].position.z   = 1e-2*zsrc[jnb]
        nb.unit[jnb].beamlets_group[0].position.phi = phi_src[jnb]
        nb.unit[jnb].beamlets_group[0].direction = 1
        nb.unit[jnb].beamlets_group[0].angle = nbi.theta_los[jnb]
        if jnb < nbi.n_box[0]:
            nb.unit[jnb].beamlets_group[0].focus.focal_length_horizontal = 6.5
            nb.unit[jnb].beamlets_group[0].focus.focal_length_vertical = 8.5
        else:
            nb.unit[jnb].beamlets_group[0].focus.focal_length_horizontal = 7.23
            nb.unit[jnb].beamlets_group[0].focus.focal_length_vertical = 11.94
        nb.unit[jnb].beamlets_group[0].divergence_component.resize(1)
# hhe following added by DPC 2019-10-24 based on data from MD
        nb.unit[jnb].beamlets_group[0].divergence_component[0].particles_fraction = 1.0
        nb.unit[jnb].beamlets_group[0].divergence_component[0].vertical = 0.01750
        nb.unit[jnb].beamlets_group[0].divergence_component[0].horizontal = 0.01750
# end of addition
        nb.unit[jnb].beamlets_group[0].beamlets.tangency_radii.resize(1)
        nb.unit[jnb].beamlets_group[0].beamlets.tangency_radii[0] = 1e-2*nbi.rtcena[jnb]

    return nb


def fill_ech(ech):

    ec = imas.ec_launchers()
    ec.ids_properties.creation_date = datetime.datetime.today().strftime("%d/%m/%y")
    ec.ids_properties.homogeneous_time = 0
    ec.ids_properties.source = os.path.abspath(__file__)

    ngy = np.sum(ech.n_gy)
    ec.launcher.resize(ngy)
    tim = np.atleast_1d(np.average(ech.t_ds))
    zero = np.array([0.], dtype=np.float64)
    for jgy in range(ngy):
        if jgy < ech.n_gy[0]:
            if ech.shot < 33725:
                ec.launcher[jgy].name = 'ECRH1'
            else:
                ec.launcher[jgy].name = 'ECRH3'
        else:
            ec.launcher[jgy].name = 'ECRH2'
        ec.launcher[jgy].identifier = 'Gy%d' %(jgy+1)
        ec.launcher[jgy].frequency.time = tim
        ec.launcher[jgy].frequency.data = np.atleast_1d(ech.freq[jgy])
        ec.launcher[jgy].power_launched.time = ech.t_ds
        ec.launcher[jgy].power_launched.data = ech.pech_ds[:, jgy]
        ec.launcher[jgy].mode.time = tim
        ec.launcher[jgy].mode.data = np.array([-1])

        ec.launcher[jgy].launching_position.time = tim
        ec.launcher[jgy].launching_position.r = np.atleast_1d(ech.R[jgy])
        ec.launcher[jgy].launching_position.z = np.atleast_1d(ech.z[jgy])
        ec.launcher[jgy].launching_position.phi = zero   # cm->m
        ec.launcher[jgy].steering_angle_pol.time = ech.t_ds
        ec.launcher[jgy].steering_angle_pol.data = np.radians(ech.the_ds[:, jgy])
        ec.launcher[jgy].steering_angle_tor.time = ech.t_ds
        ec.launcher[jgy].steering_angle_tor.data = - np.radians(ech.phi_ds[:, jgy])  # flip the sign DPC, 2020-11-24

        ec.launcher[jgy].beam.spot.size.time = tim
        logger.info('Spot size [m]   %d %8.4f', jgy, ech.width_y[jgy])
        logger.info('Curvature [1/m] %d %8.4f', jgy, ech.curv_y[jgy])
        ec.launcher[jgy].beam.spot.size.data = np.array([ech.width_y[jgy], ech.width_z[jgy]]).reshape((2, 1))
        ec.launcher[jgy].beam.spot.angle.time = tim
        ec.launcher[jgy].beam.spot.angle.data = zero

        ec.launcher[jgy].beam.phase.curvature.time = tim
        ec.launcher[jgy].beam.phase.curvature.data = np.array([1./ech.curv_y[jgy], 1./ech.curv_z[jgy]]).reshape((2, 1))
        ec.launcher[jgy].beam.phase.angle.time = tim
        ec.launcher[jgy].beam.phase.angle.data = zero

    return ec


def fill_icrf(icr):

    ic = imas.ic_antennas()

    if not hasattr(icr, 't_icp'):
        return

    strap_coord_r = np.array([2.161, 2.189, 2.208, 2.200, 2.183, 2.139, 2.094, 2.047])
    strap_coord_z = np.array([0.449, 0.307, 0.134, 0, -0.123, -0.262, -0.373, -0.454])
    ic.ids_properties.creation_date = datetime.datetime.today().strftime("%d/%m/%y")
    ic.ids_properties.homogeneous_time = 0
    ic.ids_properties.source = os.path.abspath(__file__)
    ic.ids_properties.comment = "Still iterating with EWE-2 on correctness; more work should go into the antenna description"

    nant = icr.picr.shape[1]
    ic.antenna.resize(nant)
    tim = np.atleast_1d(np.average(icr.t_ds))
    for jant in range(nant):
        ic.antenna[jant].name = str(jant+1)
        ic.antenna[jant].identifier = str(jant+1)
        ic.antenna[jant].frequency.data = np.atleast_1d(icr.freq[jant])
        ic.antenna[jant].frequency.time = tim
        ic.antenna[jant].power_launched.data = icr.picr_ds[:, jant]
        ic.antenna[jant].power_launched.time = icr.t_ds
        ic.antenna[jant].module.resize(1)
        ic.antenna[jant].module[0].name = '%s/%s' % (jant+1, 1)
        ic.antenna[jant].module[0].identifier = '%s/%s' % (jant+1, 1)
        ic.antenna[jant].module[0].frequency.data = np.atleast_1d(icr.freq[jant])
        ic.antenna[jant].module[0].frequency.time = tim
        ic.antenna[jant].module[0].power_launched.data = icr.picr_ds[:, jant]
        ic.antenna[jant].module[0].power_launched.time = icr.t_ds
        ic.antenna[jant].module[0].strap.resize(2)

        ic.antenna[jant].module[0].strap[0].width_tor = 0.18
        ic.antenna[jant].module[0].strap[0].distance_to_conductor =  0.21
        ic.antenna[jant].module[0].strap[0].phase.data = np.atleast_1d(0.0)
        ic.antenna[jant].module[0].strap[0].phase.time = tim
        ic.antenna[jant].module[0].strap[0].outline.r = strap_coord_r
        ic.antenna[jant].module[0].strap[0].outline.z = strap_coord_z
        ic.antenna[jant].module[0].strap[0].outline.phi = np.repeat(0.0, len(strap_coord_r))

        ic.antenna[jant].module[0].strap[1].width_tor = 0.18
        ic.antenna[jant].module[0].strap[1].distance_to_conductor =  0.21
        ic.antenna[jant].module[0].strap[1].phase.data = np.atleast_1d(np.radians(icr.phase[jant]))
        ic.antenna[jant].module[0].strap[1].phase.time = tim
        ic.antenna[jant].module[0].strap[1].outline.r = strap_coord_r
        ic.antenna[jant].module[0].strap[1].outline.z = strap_coord_z
        ic.antenna[jant].module[0].strap[1].outline.phi = np.repeat(0.0 + 0.095, len(strap_coord_r))

    return ic


def fill_ps(nbi, ech, icr):

    ps = imas.pulse_schedule()
    ps.ids_properties.creation_date = datetime.datetime.today().strftime("%d/%m/%y")
    ps.ids_properties.homogeneous_time = 0
    ps.ids_properties.source = os.path.abspath(__file__)

# NBI

    nb = ps.nbi

    n_nbi = np.sum(nbi.n_box)
    nb.unit.resize(n_nbi)
    tim = np.atleast_1d(np.average(nbi.t_ds))
    Rsrc = np.hypot(nbi.xsrc, nbi.ysrc)
    zsrc = nbi.xybsca
    phi_src = np.arctan2(nbi.ysrc, nbi.xsrc)

    for jnb in range(n_nbi):
        pow_frac = nbi.power_mix[jnb, :].reshape((3, 1)) # energy_mix, t
        part_frac = nbi.part_mix[jnb, :].reshape((3, 1)) # energy_mix, t
        if jnb < 4:
            nb.unit[jnb].name = 'INJ1-%s' % (jnb+1)
        else:
            nb.unit[jnb].name = 'INJ2-%s' % (jnb-3)
        nb.unit[jnb].species.a   = np.float64(nbi.abeam[jnb])
        nb.unit[jnb].species.z_n = np.float64(nbi.zbeam[jnb])
        nb.unit[jnb].species.label = spec_lbl[(nbi.zbeam[jnb], nbi.abeam[jnb])]
        nb.unit[jnb].power.reference.time = nbi.t_ds
        nb.unit[jnb].power.reference.data = nbi.pnbi_ds[:, jnb]
        nb.unit[jnb].energy.reference.time = tim
        nb.unit[jnb].energy.reference.data = np.atleast_1d(1e3*nbi.einj_kev[jnb]) #keV->eV

# ECRH
    ec = ps.ec

    ngy = np.sum(ech.n_gy)
    ec.launcher.resize(ngy)
    tim = np.atleast_1d(np.average(ech.t_ds))
    zero = np.array([0.], dtype=np.float64)
    for jgy in range(ngy):
        if jgy < ech.n_gy[0]:
            if ech.shot < 33725:
                ec.launcher[jgy].name = 'ECRH1'
            else:
                ec.launcher[jgy].name = 'ECRH3'
        else:
            ec.launcher[jgy].name = 'ECRH2'
        ec.launcher[jgy].identifier = 'Gy%d' %(jgy+1)
        ec.launcher[jgy].frequency.reference.time = tim
        ec.launcher[jgy].frequency.reference.data = np.atleast_1d(ech.freq[jgy])

        ec.launcher[jgy].power.time = ech.t_ds
        ec.launcher[jgy].power.data = ech.pech_ds[:, jgy]

        ec.launcher[jgy].steering_angle_pol.reference.time = ech.t_ds
        ec.launcher[jgy].steering_angle_pol.reference.data = np.radians(ech.the_ds[:, jgy])
        ec.launcher[jgy].steering_angle_tor.reference.time = ech.t_ds
        ec.launcher[jgy].steering_angle_tor.reference.data = - np.radians(ech.phi_ds[:, jgy])  # flip the sign DPC, 2020-11-24

    return ps


def fill_coreprof(equ, pr_obj, sig, ion_d):

    cp = imas.core_profiles()

    if hasattr(pr_obj, 'time'):
        if not hasattr(pr_obj, 'psi'):
            rhop2rhot.rhop2rhot(equ, pr_obj)
        cp.time = pr_obj.time
        if hasattr(sig, 'zef'):
            ion_d['zef1d'] = np.interp(cp.time, sig.zef.time, sig.zef.data)
        else:
            ion_d['zef1d'] = cp.time*0 + 1.5      # in the absence of any Zeff data, assume 1.5
        logger.warning('Zeff forced to 1.5 in the absence of any other data')

    nt_cp = len(cp.time)
    nx_cp = len(pr_obj.rho)
 
    nt_equ, nx_equ = equ.rho_tor.shape

    rho_tor_n = pr_obj.rho
    sgn_psi = -1 # COCO 17 to 11

    cp.profiles_1d.resize(nt_cp)

    cp.ids_properties.homogeneous_time = 1   # same timebase for core_profiles IDS
    cp.ids_properties.creation_date = datetime.datetime.today().strftime("%d/%m/%y")
    cp.ids_properties.source = os.path.abspath(__file__)

    eV_m3_to_Pa = 1.602*1e-19

# Interpolate equilibrium quantities on profiles' time grid

    tfb = np.interp(cp.time, equ.time, equ.tfl[:, -1])
    B0  = np.interp(cp.time, equ.time, equ.B0)
    rho_tor_b = np.sqrt(tfb/(np.pi*B0))
    rho_tor = np.outer(rho_tor_b, rho_tor_n) # (nt_cp)x(nx_cp)

    tfln_cpt = np.zeros([nt_cp, nx_equ])
    vol_cpt  = np.zeros([nt_cp, nx_equ])
    area_cpt = np.zeros([nt_cp, nx_equ])
    q_cpt    = np.zeros([nt_cp, nx_equ])
    for jrho in range(nx_equ):
        tfln_cpt[:, jrho] = np.interp(cp.time, equ.time, equ.tfln[:, jrho])
        area_cpt[:, jrho] = np.interp(cp.time, equ.time, equ.area[:, jrho])
        vol_cpt [:, jrho] = np.interp(cp.time, equ.time, equ.vol [:, jrho])
        q_cpt   [:, jrho] = np.interp(cp.time, equ.time, equ.q   [:, jrho])

    zimp = np.float64(ion_d['Zimp'])
    if 'I' not in pr_obj.prof.keys():
        logger.warning('No Ti data --- setting Ti = Te')
    if 'V' not in pr_obj.prof.keys():
        logger.warning('No vtor data --- setting vtor = 0')

    for jt in range(nt_cp):
        te   = pr_obj.prof['E'].yfit_rhot[jt, :]
        ne   = pr_obj.prof['N'].yfit_rhot[jt, :]
        nimp = (ion_d['zef1d'][jt] - 1.)/(zimp**2 - zimp)*ne
        nd = ne - zimp*nimp
        if 'I' in pr_obj.prof.keys():
            ti = pr_obj.prof['I'].yfit_rhot[jt, :]
        else:
            ti = te

# Map area, vol to profiles# rho-grid
        area = np.interp(rho_tor_n, tfln_cpt[jt, :], area_cpt[jt, :])
        vol  = np.interp(rho_tor_n, tfln_cpt[jt, :],  vol_cpt[jt, :])
        q    = np.interp(rho_tor_n, tfln_cpt[jt, :],    q_cpt[jt, :])
# Filling

        cp.profiles_1d[jt].grid.rho_tor      = rho_tor[jt, :]  
        cp.profiles_1d[jt].grid.psi          = pr_obj.psi[jt, :]*sgn_psi
        cp.profiles_1d[jt].grid.rho_tor_norm = rho_tor_n
        cp.profiles_1d[jt].grid.area         = area
        cp.profiles_1d[jt].grid.volume       = vol
        cp.profiles_1d[jt].electrons.temperature     = te
        cp.profiles_1d[jt].electrons.density         = ne
        cp.profiles_1d[jt].electrons.density_thermal = ne
        cp.profiles_1d[jt].electrons.pressure = eV_m3_to_Pa*ne*te
        cp.profiles_1d[jt].electrons.pressure_thermal = eV_m3_to_Pa*ne*te

        cp.profiles_1d[jt].q = q

        cp.profiles_1d[jt].t_i_average = ti
        cp.profiles_1d[jt].ion.resize(2)

        cp.profiles_1d[jt].ion[0].z_ion = np.float64(ion_d['Zmain'])
        cp.profiles_1d[jt].ion[0].temperature = ti
        cp.profiles_1d[jt].ion[0].label       = spec_lbl[(ion_d['Zmain'], ion_d['Amain'])]
        if 'V' in  pr_obj.prof.keys():
            vt = pr_obj.prof['V'].yfit_rhot[jt, :]
            if pr_obj.prof['V'].units == 'rad/s':
                cp.profiles_1d[jt].ion[0].rotation_frequency_tor = vt
            else:
                cp.profiles_1d[jt].ion[0].velocity_tor = vt
        else:
            vt = np.zeros_like(te)
            cp.profiles_1d[jt].ion[0].rotation_frequency_tor = vt
        cp.profiles_1d[jt].ion[0].density          = nd
        cp.profiles_1d[jt].ion[0].density_thermal  = nd
        cp.profiles_1d[jt].ion[0].pressure         = eV_m3_to_Pa*nd*ti
        cp.profiles_1d[jt].ion[0].pressure_thermal = eV_m3_to_Pa*nd*ti
        cp.profiles_1d[jt].ion[0].element.resize(1)
        cp.profiles_1d[jt].ion[0].element[0].a   = np.float64(ion_d['Amain'])
        cp.profiles_1d[jt].ion[0].element[0].z_n = np.float64(ion_d['Zmain'])
        cp.profiles_1d[jt].ion[0].element[0].atoms_n = 1
        cp.profiles_1d[jt].ion[0].multiple_states_flag = 0
    
        cp.profiles_1d[jt].ion[1].z_ion = zimp
        cp.profiles_1d[jt].ion[1].temperature = ti
        cp.profiles_1d[jt].ion[1].label = spec_lbl[(ion_d['Zimp'], ion_d['Aimp'])]
        if 'V' in pr_obj.prof.keys():
            vt = pr_obj.prof['V'].yfit_rhot[jt, :]
            cp.profiles_1d[jt].ion[0].velocity_tor = vt
        elif 'ANGF' in pr_obj.prof.keys():
            angf = pr_obj.prof['ANGF'].yfit_rhot[jt, :]
            cp.profiles_1d[jt].ion[0].rotation_frequency_tor = angf
        else:
            cp.profiles_1d[jt].ion[0].rotation_frequency_tor = np.zeros_like(te)
        cp.profiles_1d[jt].ion[1].density          = nimp
        cp.profiles_1d[jt].ion[1].density_thermal  = nimp
        cp.profiles_1d[jt].ion[1].pressure         = eV_m3_to_Pa*nimp*ti
        cp.profiles_1d[jt].ion[1].pressure_thermal = eV_m3_to_Pa*nimp*ti
        cp.profiles_1d[jt].ion[1].element.resize(1)
        cp.profiles_1d[jt].ion[1].element[0].a   = np.float64(ion_d['Aimp'])
        cp.profiles_1d[jt].ion[1].element[0].z_n = zimp
        cp.profiles_1d[jt].ion[1].element[0].atoms_n = 1
        cp.profiles_1d[jt].ion[1].multiple_states_flag = 0
    
        cp.profiles_1d[jt].zeff = np.repeat(ion_d['zef1d'][jt], nx_cp)

    return cp


def fill_equ(equ, psiN_cut=0.994):

    equ.to_coco(cocos_out=11)

    nt_eq, nrho = equ.tfl.shape

    eq = imas.equilibrium()

    eq.code.name = "trview"
    eq.code.version = "2020.04.04"
    eq.ids_properties.source = "ASTRA equil"
    eq.ids_properties.provider = os.getenv('USER')
    eq.ids_properties.creation_date = datetime.datetime.today().strftime("%d/%m/%y")

    eq.time_slice.resize(nt_eq)
    eq.ids_properties.homogeneous_time = 1

    eq.time = np.array(equ.time)
    eq.vacuum_toroidal_field.r0 = equ.R0
    eq.vacuum_toroidal_field.b0 = np.array(equ.B0)

    Rmesh = np.array(equ.Rmesh, dtype=np.float32)
    Zmesh = np.array(equ.Zmesh, dtype=np.float32)
    pol_flux2d = np.array(equ.pfm, dtype=np.float32)

# B components
    br, bz, bt = sf.Bmesh(equ)
    br = np.array(br)
    bz = np.array(bz)
    bt = np.array(bt)

    prof_map = {'psi': 'pfl', 'phi': 'tfl', \
        'rho_tor': 'rho_tor', 'rho_tor_norm': 'rho_tor_n', \
        'pressure': 'pres', 'dpressure_dpsi': 'dpres', \
        'f_df_dpsi': 'ffp', 'q': 'q', 'j_tor': 'jav', \
        'volume': 'vol', 'dvolume_dpsi': 'dvol', \
        'area': 'area', 'darea_psi': 'darea'}

    for itim in range(nt_eq):
 
        eq.time_slice[itim].boundary_separatrix.outline.r = equ.geofit.r_bnd[itim, :-1]
        eq.time_slice[itim].boundary_separatrix.outline.z = equ.geofit.z_bnd[itim, :-1]
        eq.time_slice[itim].profiles_1d.f = 2e-7*np.array(equ.jpol[itim])
        for imas_lbl, sf_lbl in prof_map.items():
            eq.time_slice[itim].profiles_1d.__dict__[imas_lbl] = \
                np.array(equ.__dict__[sf_lbl][itim])

# global quantities
# this all assumes standard AUG shotfile writing starting from separatrix and moving towards axis
# This should be checked in future, as, I believe, IDE does not do this...
        eq.time_slice[itim].global_quantities.volume = np.double(equ.vol[itim, -1])
        eq.time_slice[itim].global_quantities.area   = np.double(equ.area[itim, -1])
        eq.time_slice[itim].global_quantities.psi_axis = np.double(equ.psi0[itim])
        eq.time_slice[itim].global_quantities.psi_boundary = np.double(equ.psix[itim])
        eq.time_slice[itim].global_quantities.ip = np.double(equ.ip[itim])
        eq.time_slice[itim].global_quantities.magnetic_axis.r = np.double(equ.Rmag[itim])
        eq.time_slice[itim].global_quantities.magnetic_axis.z = np.double(equ.Zmag[itim])
        eq.time_slice[itim].global_quantities.magnetic_axis.b_field_tor = np.double(equ.B0[itim])

# now get flux value near to separatrix (psiN = pst_cut) for contouring
        rsurf = equ.geofit.rscat[itim][0][:-1]
        zsurf = equ.geofit.zscat[itim][0][:-1]

# get some basic quantities
        rin  = np.min(rsurf)
        rout = np.max(rsurf)
        jz_top = np.argmax(zsurf)
        jz_bot = np.argmin(zsurf)
        ztop  = zsurf[jz_top]
        rztop = rsurf[jz_top]
        zbot  = zsurf[jz_bot]
        rzbot = rsurf[jz_bot]

        rgeo = (rin + rout) / 2.
        aminor = (rout - rin) / 2.
        tria_upp = (rgeo - rztop ) /aminor 
        tria_low = (rgeo - rzbot ) /aminor 
        tria = (tria_upp + tria_low) / 2.

        zgeo = (zbot + ztop) / 2.
        elong_up = (ztop - zgeo) / aminor
        elong_low = (zgeo - zbot) / aminor
        elong = (ztop - zbot) / 2. / aminor

# now step through contour with theta to get more regular boundary representation

        eq.time_slice[itim].boundary.psi_norm = psiN_cut
        eq.time_slice[itim].boundary.psi = np.interp(psiN_cut, equ.psiN[itim], equ.pfl[itim])
        eq.time_slice[itim].boundary.outline.r = rsurf
        eq.time_slice[itim].boundary.outline.z = zsurf
        eq.time_slice[itim].boundary.geometric_axis.r = rgeo
        eq.time_slice[itim].boundary.geometric_axis.z = zgeo
        eq.time_slice[itim].boundary.minor_radius = aminor
        eq.time_slice[itim].boundary.elongation = elong
        eq.time_slice[itim].boundary.elongation_upper = elong_up
        eq.time_slice[itim].boundary.elongation_lower = elong_low
        eq.time_slice[itim].boundary.triangularity = tria
        eq.time_slice[itim].boundary.triangularity_lower = tria_low
        eq.time_slice[itim].boundary.triangularity_upper = tria_upp

# 2D matrix
        eq.time_slice[itim].profiles_2d.resize(1)
        eq.time_slice[itim].profiles_2d[0].grid_type.name = 'rectangular'
        eq.time_slice[itim].profiles_2d[0].grid_type.description = identifiers.poloidal_plane_coordinates_identifier.poloidal_plane_coordinates_identifier['rectangular']['description']
        eq.time_slice[itim].profiles_2d[0].grid_type.index = identifiers.poloidal_plane_coordinates_identifier.poloidal_plane_coordinates_identifier['rectangular']['index']

        eq.time_slice[itim].profiles_2d[0].grid.dim1 = Rmesh
        eq.time_slice[itim].profiles_2d[0].grid.dim2 = Zmesh
        eq.time_slice[itim].profiles_2d[0].psi = pol_flux2d[:, :, itim]
        eq.time_slice[itim].profiles_2d[0].b_field_r   = br[itim, :, :]
        eq.time_slice[itim].profiles_2d[0].b_field_z   = bz[itim, :, :]
        eq.time_slice[itim].profiles_2d[0].b_field_tor = bt[itim, :, :]

    return eq
